//---------------------------------------------------------------------------

#pragma hdrstop

#include "dictionary.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)



Dictionary::  Dictionary( bool (*KeyEqual)(void *key1, void *key2),
             THashKey (*KeyHashCode)(void *key),
             void (*WriteElement)(HNode *node, FILE *f),
             bool (*ReadElement)(HNode *node, FILE *f)
             ){
    memset(this,0,sizeof(Dictionary));
    this->KeyEqual = KeyEqual;
    this->KeyHashCode = KeyHashCode;
    this->WriteElement = WriteElement;
    this->ReadElement = ReadElement;
    this->last = &this->list;
}


Dictionary::  Dictionary(){
    memset(this,0,sizeof(Dictionary));
    this->KeyEqual = KeyEqualDefault;
    this->KeyHashCode = KeyHashCodeDefault;
    this->WriteElement = WriteElementDefault;
    this->ReadElement = ReadElementDefault;
    this->last = &this->list;
}

Dictionary::  ~Dictionary(){ //      
   HNode *p, *tmp;

   p = list;
   while( p ){
     tmp = p;
     p = p->nextAll;
     delete tmp;
   }
}



HNode* Dictionary::Look(void *key){
    THashKey key_hash_code = KeyHashCode(key);
    HNode *p = tbl[key_hash_code & (H_TBL_CNT-1)];
    for(; p; p = p->nextSub)
     if( p->key_hash_code == key_hash_code &&
         KeyEqual(p->key, key) )
         return p;
    return NULL;
}

HNode* Dictionary::Add(void *key, void *value){
  HNode *node = new HNode(key,value,KeyHashCode(key),cnt);
  if( !node ) return NULL;

  HNode **entry = Entry(key); //   
  node->nextSub = *entry;
  *entry = node;

  *last = node;  //     
  last = &(node->nextAll);

                 //-  
  HNode **index_entry = &tbl_index[cnt & 31];
  node->nextIndex = *index_entry;
  *index_entry = node;


  cnt++;
  return node;
}

HNode* Dictionary::GetElementAt(int index){
  for( HNode* p = tbl_index[(unsigned)index&31]; p; p = p->nextIndex)
   if(p->index == index) return p;
  return NULL;
}

bool Dictionary::Load(char fname[]){
  FILE *f = fopen(fname,"rb");
  if(!f) return false;
  HNode tmp(0,0,0,0);

  while( ReadElement(&tmp,f)  &&
         Add( tmp.key, tmp.value ));

  fclose(f);
  return true;
}

bool Dictionary::Save(char fname[]){
  FILE *f = fopen(fname,"wb");
  if(!f) return false;
  for(HNode *p = list; p; p = p->nextAll)
    WriteElement(p,f);
  fclose(f);  
  return true;
}




//    
bool KeyEqualDefault(void *key1, void *key2){

   return  strcmp((char*)key1,(char*)key2) == 0;
}

THashKey KeyHashCodeDefault(void *key){
   return GetStrHashKey((char*)key);

}

void WriteElementDefault(HNode *node, FILE *f){

  fprintf(f,"%s=%s\n", node->key, node->value);
}

bool ReadElementDefault(HNode *node, FILE *f){
  const int buf_size = 1024;
  char buf[buf_size];

  do{

    if( fgets(buf,buf_size-1,f) == NULL ) return false;
    Trim(buf);

  }while(buf[0] == '#' || buf[0] == '\0'); //skip comment

  char *p = strchr(buf,'=');
  if(p == NULL) return false; //p = buf + strlen(buf);  // end

  *p = '\0';

  node->key = strdup( Trim(buf) );
  node->value = strdup( Trim(p+1) );

  return  node->key != NULL &&  node->value != NULL;
}




